---------------------------------------------------------------
-- Commandos 2 ABI Export utility, v1.0 (1st December 2015)  --
-- by herbert3000 (ferdinand.graf.zeppelin@gmail.com)        --
---------------------------------------------------------------

/*
	Info:
	all comments starting with -- ABI:  relate to the Commandos 2 ABI File Format Document (Src/Doc/Comm2_ABI.en.htm)
	arranged by NeoRAGEx2002 and translated by F.R.C. which is included in the Commandos Developing Toolkit
*/

-- A simple stack
struct myStack (
	top = 0,
	array = #(),
	function push val = (
		top += 1
		array[top] = val
	),
	function pop = (
		if top > 0 then (
			top -= 1
			return array[top+1]
		) else (
			print "Error: stack underflow"
			return undefined
		)
	)
)

function getRoot node = (
	if isvalidnode node do 
	(
		while node.parent != undefined do node = node.parent
		node
	)
)

-- correct version of writeString, MAXScript also prints the null-character
function writeString_ file string =
(
	for i = 1 to string.count do
	(
		writeByte file (bit.charAsInt string[i])
	)
)

-- write a 0-padded string
function writeStringPadded file string totalLength =
(
	writeString file string
	for i = string.count to totalLength-2 do
	(
		writeByte file 0
	)
)

function writePalette file palette =	-- ABI: Texture Info DB.Palettte DS
(
	for i = 1 to palette.count do
	(
		color = palette[i]
		writeByte file (color.r + 1)	-- ABI: Red
		writeByte file (color.g + 1)	-- ABI: Green
		writeByte file (color.b + 1)	-- ABI: Blue
	)
)

function writePixels file tex =		-- ABI: Texture Info DB.Indexed Bitmap DS
(
	for h = 1 to tex.height do
	(
		pixels = getIndexedPixels tex [0, h-1] tex.width
		for w = 1 to tex.width do
		(
			writeByte file pixels[w]	-- ABI: Bitmap Data
		)
	)
)

function writeBitmap file tex =	-- ABI: Texture Info DB.Texture Info DS
(
	writeLong file 0			-- ABI: Block Identifier
	writeLong file tex.width	-- ABI: Width
	writeLong file tex.height	-- ABI: Height
	
	tokens = filterString tex.filename "\\"	-- split filename
	name = tokens[tokens.count] -- filename without path
	writeStringPadded file name 32	-- ABI: Name 
	
	writePalette file tex.palette
	writePixels file tex
)

rollout exportbutton "Commandos ABI Exporter" width:200 height:125
(
	label lbl1 "Commandos ABI Exporter" pos:[42,10]
	button btn_export "Export as ABI" pos:[30,35] width:145 height:45
	checkBox bonesHaveRoot "Bones have Root" pos:[30,108] checked:false
	checkBox writeParents "Output Bone Hierarchy" pos:[30,128] checked:false
	label lbl2 "v1.0                       by herbert3000" pos:[18,160]
	
	on btn_export pressed do
	(
	
	filename = getSaveFileName types:"Commandos ABI file|*.abi|All|*.*|"
	if filename != undefined then
	(
	file = fopen filename "wb"
	
	--clearListener()
	
	-- Define constants
	identifier = "LDMB"
	version = "1050"
	
	numTimeAxis = 0
	numAnimations = 1
	numModels = 1
	numBones = 0
	textures = #()
	
	-- Get BMP Filenames --
	materials = getClassInstances Multimaterial
	
	for mat in materials do
	(
		numSubs = mat.numSubs
		for s in 1 to numSubs do
		(
			bitmap = mat.materialList[s].diffuseMap.bitmap
			append textures bitmap
		)
	)
	
	-- count bones
	b = for o in Objects where (classOf o == BoneGeometry) collect o
	numBones = b.count
	numTimeAxis = numBones
	
	-- ABI: Basic Info DB (Header)
	writeString_ file identifier	-- ABI: Identifier
	writeString_ file version		-- ABI: Version
	
	writeLong file numTimeAxis		-- ABI: Num TimeAxis
	writeLong file numAnimations	-- ABI: Num Animation
	
	writeLong file textures.count	-- ABI: Num Texture
	for t in textures do
	(
		writeBitmap file t
	)
	
	writeLong file numModels		-- ABI: Num Model
	writeLong file numBones			-- ABI: Num Bone
	
	names    = #()
	verts    = #()
	faces    = #()
	uvCoords = #()
	matIDs   = #()
	bones    = #()
	vertOff  = #() -- vertices offset counter
	offsets  = #()
	
	
	objStack = myStack()
	idStack = myStack()
	parentIds = #()
	
	parentId = -1
	
	if bonesHaveRoot.checked then
	(
		root = getRoot objects[1]
		objStack.push root
		idStack.push parentId
	)
	else
	(
		for o in Objects where (isGroupHead o == true) do
		(
			objStack.push o
			idStack.push parentId
		)
	)
	
	while objStack.top > 0 do
	(
		obj = objStack.pop()
		
		id = idStack.pop()
		append parentIds id
		
		parentId += 1
		
		append names obj.name
		
		append vertOff verts.count
		offset = verts.count
		
		for c in obj.children do
		(
			if (classOf c == Editable_Poly) then	-- Polygon
			(
				numVerts = polyop.getNumVerts c
				for j = 1 to numVerts do	-- Vertices
				(
					vert = polyOp.getVert c j
					append verts vert
				)
				
				numFaces = polyop.getNumFaces c
				for j = 1 to numFaces do	-- Faces
				(
					face = polyOp.getFaceVerts c j
					matID = polyOp.getFaceMatID c j
					append faces face
					append matIDs matID
					append offsets offset
				)
				for f in (c.faces as bitarray) do
				(
					for v in (polyOp.getMapFace c 1 f) do
					(
						uv = polyOp.getMapVert c 1 v
						append uvCoords uv
					)
				)
			)
			else if (classOf c == BoneGeometry ) then	-- Bone
			(
				append bones c
			)
			else if (isGroupHead c) do 	-- Group
			(
				objStack.push c
				idStack.push parentId
			)
		)
	)
	
	append vertOff verts.count
	
	-- ABI: Model DB
	writeLong file verts.count		-- ABI: Num Vertex
	writeLong file faces.count		-- ABI: Num Polygon
	name = maxFileName	-- filename of opened .max file
	writeStringPadded file maxFileName 32	-- ABI: Name
	
	for v in verts do		-- ABI: Vertices DS
	(
		writeFloat file v.x	-- ABI: x
		writeFloat file v.y	-- ABI: y
		writeFloat file v.z	-- ABI: z
		format "v % % %\n" v.x v.y v.z
	)
	
	vertexCounter = 0
	
	for i in 1 to faces.count do (	-- ABI: Polygon DS
		f = faces[i]
		writeByte file f.count		-- ABI: n
		
		writeByte file (matIDs[i] - 1)	-- ABI: Texture ID
		format "\nf"
		for j = 1 to f.count do (
			index = f.count - j + 1 -- reverse order
			vertIndex = f[index] + offsets[i]
			writeShort file (vertIndex - 1)		-- ABI: Vertex ID
			writeShort file (uvCoords[vertexCounter + index].x * 4096)	-- ABI: U
			writeShort file (uvCoords[vertexCounter + index].y * -4096)	-- ABI: V
			
			format " %" vertIndex
		)
		vertexCounter += f.count
	)
	
	for i in 1 to bones.count do	-- ABI: Vidx to Bone DS
	(
		writeLong file vertOff[i]	-- ABI: Start Vertex Index
		writeLong file vertOff[i+1]	-- ABI: End Vertex Index
	)
	
	for i in 1 to bones.count do	-- ABI: Bone Hierarchy DB
	(
		bonePos = bones[i].pos
		if bonesHaveRoot.checked and writeParents.checked then
		(
			writeLong file parentIds[i]	-- ABI: Parent Index
		)
		else
		(
			writeLong file -1 			-- means no parent
		)
		writeFloat file bonePos.x	-- ABI: Global Offset Vector
		writeFloat file bonePos.y
		writeFloat file bonePos.z
		writeStringPadded file names[i] 32	-- ABI: Node Name
		writeFloat file 0			-- ABI: UNKNOWN
	)
	
	-- ABI: Animation Chunk
	for i in 1 to numAnimations do
	(
		for j in 1 to numTimeAxis do
		(
			writeLong file 1	-- ABI: num_translateKeyframes
			writeByte file 0	-- ABI: timestamp
			writeShort file 0	-- ABI: dx
			writeShort file 0	-- ABI: dy
			writeShort file 0	-- ABI: dz
			
			writeLong file 1	-- ABI: num_rotateKeyframes
			writeByte file 0	-- ABI: timestamp
			writeShort file 0	-- ABI: x
			writeShort file 0	-- ABI: y
			writeShort file 0	-- ABI: z
			writeShort file 32767	-- ABI: w
		)
		
		writeStringPadded file "base" 32	-- ABI: animation_name
		writeLong file 1	-- unknown
		writeLong file -1
		writeLong file -1
		writeLong file 0
		writeLong file 0
		writeLong file 0
		writeLong file 0
		
		writeLong file numBones	-- ABI: num_related_bones
		for j in 1 to numBones do
		(
			writeLong file (j - 1)	-- ABI: bone_id
			writeLong file (j - 1)	-- ABI: transform_time_axis_idx
		)
	)
	
	fclose file
	
	) --end "if filename != undefined then"
	
	-- Show the MAXScript Listener Window
	--actionMan.executeAction 0 "40472"
	
	) --end "on btn_export pressed do"

)
CreateDialog exportbutton 202 184 25 25